/*
energy.c
Author: Hayden Walles
Date: 9 September 2007
Useful functions for implementing energy computations.
*/

/*
    Copyright (C) 2007, 2008, 2009 Hayden Walles

    This file is part of Seamstress.

    Seamstress is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Seamstress is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "seams.h"


/*Alters the automatically computed energy function according to the user's annotations.*/
void modulateEnergy(SEAM_COMMON *pic,int transposed){
  int i,j;
  int mark;
  if((pic->energy==NULL)||(pic->marks==NULL))
    return;
  //  transposed = pic->width != matrixCols(pic->marks);
  
  for(i=0;i<pic->height;i++){
    for(j=0;j<pic->width;j++){
      if(!transposed)
	mark=getByte(pic->marks,i,j);
      else
	mark=getByte(pic->marks,j,i);
      
      if(mark==1)
	setInteger(pic->energy,i,j,255);
      else if(mark==2)
	setInteger(pic->energy,i,j,0);
      
    }
  }


}


void remodulateEnergyVertically(SEAM_UNPICKER *pic,int *froms, int *tos){
  int i,j;
  int transposed=pic->direction!=0;
  int mark;
  if((pic->common.energy==NULL)||(pic->common.marks==NULL))
    return;

  //transposed = pic->common.width!=matrixCols(pic->common.marks);
 
  for(i=0;i<pic->common.height;i++){
    for(j=froms[i];j<=tos[i];j++){
      if(!transposed)
	mark=getByte(pic->common.marks,i,pic->nodes[i*pic->common.width+j].offset);
      else
	mark=getByte(pic->common.marks,pic->nodes[i*pic->common.width+j].offset,i);
      
      if(mark==1)
	setInteger(pic->common.energy,i,pic->nodes[i*pic->common.width+j].offset,255);
      else if(mark==2)
	setInteger(pic->common.energy,i,pic->nodes[i*pic->common.width+j].offset,0);
      
    }
  }


}


#if 0
/*This is experimental and isn't properly supported yet.*/
void remodulateEnergyHorizontally(SEAM_COMMON *pic,int *froms, int *tos){
  int i,j;
  int mark,transposed;

  if((pic->energy==NULL)||(pic->marks==NULL))
    return;

  transposed = pic->width!=matrixCols(pic->marks);

  for(j=0;j<pic->width;j++){
    for(i=froms[j];i<=tos[j];i++){
      if(!transposed)
	mark=getByte(pic->marks,i,j);
      else
	mark=getByte(pic->marks,j,i);
      
      if(mark==1)
	setInteger(pic->energy,i,j,255);
      else if(mark==2)
	setInteger(pic->energy,i,j,0);
      
    }
  }

}
#endif


int unpickerComputeEnergy(SEAM_UNPICKER *pic,SEAM_MARKS *marks,void (*updater)(int done,int of,void *user),void *user,int *error){
  
  if(marks!=NULL){
    if((matrixCols(marks)!=pic->actualwidth)||(matrixRows(marks)!=pic->actualheight)){
      *error=SEAMSTRESS_E_BADMARKS;
      return 0;
    }
    pic->common.marks=marks;
  }
  
  if(!computeGaussianGradientEnergy(&pic->common,updater,user,error))
    return 0;
  
  modulateEnergy(&pic->common,pic->direction!=0);
  SET_ERROR(SEAMSTRESS_E_OK);
  return 1;

}


int unweaverComputeEnergy(SEAM_UNWEAVER *pic,SEAM_MARKS *marks,void (*updater)(int done,int of,void *user),void *user,int *error){
  
  if(marks!=NULL){
    if((matrixCols(marks)!=pic->common.width)||(matrixRows(marks)!=pic->common.height)){
      *error=SEAMSTRESS_E_BADMARKS;
      return 0;
    }
    pic->common.marks=marks;
  }
  
  if(!computeGaussianGradientEnergy(&pic->common,updater,user,error))
    return 0;
  
  modulateEnergy(&pic->common,0);

  SET_ERROR(SEAMSTRESS_E_OK);
  return 1;

}


/*It would be nice to extend this ability to unweavers.*/
int adjustEnergyVertically(SEAM_UNPICKER *pic,int hremove,int vremove,int *seam){
  int i,fi;
  int width=pic->common.width-hremove;
  int height=pic->common.height-vremove;
  int b,t;
  int l,r;
  /*Work out which pixels will need to be recomputed via f1, and f2, then recompute them.*/
  int w,h;
  int error;
  int hborder,vborder;
#ifdef HAVE_DYNAMIC_ARRAYS
  int ls[height],rs[height];
#elif defined(HAVE_ALLOCA)
  int *ls,*rs;
  ls=(int *)alloca(sizeof(int)*height);
  rs=(int *)alloca(sizeof(int)*height);
#else
  /*I should replace this with heap-allocated arrays, of course.*/
#error Need dynamic arrays or alloca!
#endif


  hborder=(matrixCols(pic->common.luminance)-pic->common.width)/2;
  vborder=(matrixRows(pic->common.luminance)-pic->common.height)/2;


  removeMatrixVerticalSeam(pic->common.luminance,width,height,hborder,vborder,seam);
  
  w=matrixCols(pic->common.luminance)-pic->common.width;
  h=matrixRows(pic->common.luminance)-pic->common.height;
      
  b=-(h/2);
  t=b+h-1;

  l=MAX(0,seam[0]-w+1);
  r=MIN(width-1,seam[0]);
  for(fi=0;fi<=t;fi++){
    ls[fi]=l;
    rs[fi]=r;
  }

  for(i=1;i<height;i++){
    l=MAX(0,seam[i]-w+1);
    r=MIN(width-1,seam[i]);
    for(fi=b;fi<t;fi++){
      if((i+fi>0)&&(i+fi<height)){
	ls[i+fi]=MIN(l,ls[i+fi]);
	rs[i+fi]=MAX(r,rs[i+fi]);
      }
    }
    if((i+fi>0)&&(i+fi<height)){
      ls[i+fi]=l;
      rs[i+fi]=r;
    }
  }

  recomputeGaussianGradientEnergy(pic,ls,rs);
  remodulateEnergyVertically(pic,ls,rs);

  /*Finallu, make sure that partialMinimalSeams knows which energy values have been altered in the first row.*/
  pic->lastfrom=ls[0];
  pic->lastto=rs[0];

  return 1;


}

#if 0
/*This is experimental, and isn't properly supported yet.*/
int adjustEnergyHorizontally(SEAM_COMMON *pic,int hremove,int vremove,int *seam){
  int i,fi;
  int width=pic->width-hremove;
  int height=pic->height-vremove;
  int ls[width],rs[width];
  int b,t;
  int l,r;
  /*Work out which pixels will need to be recomputed via f1, and f2, then recompute them.*/
  int w,h;
  int error;
  int hborder,vborder;

  hborder=(matrixCols(pic->luminance)-pic->width)/2;
  vborder=(matrixRows(pic->luminance)-pic->height)/2;

   removeMatrixHorizontalSeam(pic->luminance,width,height,hborder,vborder,seam);
  
  w=matrixCols(pic->luminance)-pic->width;
  h=matrixRows(pic->luminance)-pic->height;
  
  b=-(w/2);
  t=b+w-1;

  l=MAX(0,seam[0]-h+1);
  r=MIN(height-1,seam[0]);
  for(fi=0;fi<=t;fi++){
    ls[fi]=l;
    rs[fi]=r;
  }

  for(i=1;i<width;i++){
    l=MAX(0,seam[i]-h+1);
    r=MIN(height-1,seam[i]);
    for(fi=b;fi<t;fi++){
      if((i+fi>0)&&(i+fi<width)){
	ls[i+fi]=MIN(l,ls[i+fi]);
	rs[i+fi]=MAX(r,rs[i+fi]);
      }
    }
    if((i+fi>0)&&(i+fi<width)){
      ls[i+fi]=l;
      rs[i+fi]=r;
    }
  }

  recomputeGaussianGradientEnergy(pic,ls,rs);
  remodulateEnergyHorizontally(pic,ls,rs);
  return 1;


}

#endif



#if 0
int removeMatrixVerticalSeam(SEAM_UNPICKER *pic,MATRIX *target, int *seam){
  int i,j;
  int *rowptr;
  int vborder,hborder;
  int width=pic->width-pic->seamcount; //this is the _new_ width

  vborder=(matrixRows(target)-pic->height)/2;
  hborder=(matrixCols(target)-pic->width)/2;
  
  rowptr = target->buffer.integers;
  
  if(seam[0]==0){
    for(j=0;j<hborder;j++)
      rowptr[j]=rowptr[hborder+1];
    for(j=hborder;j<hborder+width;j++)
      rowptr[j]=rowptr[j+1];
  }
  else if(seam[0]==width){
    for(j=hborder+width-1;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-1];
  }
  else {
    for(j=seam[0]+hborder;j<=width+hborder;j++)
      rowptr[j]=rowptr[j+1];
  }
  
  rowptr+=matrixCols(target);
  for(i=1;i<vborder+1;i++){
    for(j=0;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-matrixCols(target)];
    rowptr+=matrixCols(target);
  }

  for(i=1;i<pic->height;i++){
    
    if(seam[i]==0){
      for(j=0;j<hborder;j++)
	rowptr[j]=rowptr[hborder+1];
      for(j=hborder;j<hborder+width;j++)
	rowptr[j]=rowptr[j+1];
    }
    else if(seam[i]==width){
      for(j=hborder+width-1;j<hborder+width+hborder;j++)
	rowptr[j]=rowptr[j-1];
    }
    else {
      for(j=seam[i]+hborder;j<=width+hborder;j++)
	rowptr[j]=rowptr[j+1];
    }
    rowptr+=matrixCols(target);
  }

  for(i=0;i<vborder;i++){
    for(j=0;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-matrixCols(target)];
    rowptr+=matrixCols(target);
  }

  return 1;

}

#endif

int removeMatrixVerticalSeam(MATRIX *target,int width,int height,int hborder,int vborder, int *seam){
  int i,j;
  int *rowptr;
  /*width and height are the current effective widths - the matrix may actually be larger.*/

  rowptr = target->buffer.integers;
  
  if(seam[0]==0){
    for(j=0;j<hborder;j++)
      rowptr[j]=rowptr[hborder+1];
    for(j=hborder;j<hborder+width;j++)
      rowptr[j]=rowptr[j+1];
  }
  else if(seam[0]==width){
    for(j=hborder+width-1;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-1];
  }
  else {
    for(j=seam[0]+hborder;j<=width+hborder;j++)
      rowptr[j]=rowptr[j+1];
  }
  
  rowptr+=matrixCols(target);
  for(i=1;i<vborder+1;i++){
    for(j=0;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-matrixCols(target)];
    rowptr+=matrixCols(target);
  }

  for(i=1;i<height;i++){
    
    if(seam[i]==0){
      for(j=0;j<hborder;j++)
	rowptr[j]=rowptr[hborder+1];
      for(j=hborder;j<hborder+width;j++)
	rowptr[j]=rowptr[j+1];
    }
    else if(seam[i]==width){
      for(j=hborder+width-1;j<hborder+width+hborder;j++)
	rowptr[j]=rowptr[j-1];
    }
    else {
      for(j=seam[i]+hborder;j<=width+hborder;j++)
	rowptr[j]=rowptr[j+1];
    }
    rowptr+=matrixCols(target);
  }

  for(i=0;i<vborder;i++){
    for(j=0;j<hborder+width+hborder;j++)
      rowptr[j]=rowptr[j-matrixCols(target)];
    rowptr+=matrixCols(target);
  }

  return 1;

}

#if 0
/*This is experimental and isn't properly supported yet.*/
int removeMatrixHorizontalSeam(MATRIX *target,int width,int height,int hborder,int vborder, int *seam){
  int i,j;
  /*width and height are the current effective widths - the matrix may actually be larger.*/

  if(seam[0]==0){
    for(i=0;i<vborder;i++)
      setInteger(target,i,0,getInteger(target,vborder+1,0));
    for(i=vborder;i<vborder+height;i++)
      setInteger(target,i,0,getInteger(target,i+1,0));
  }
  else if(seam[0]==height){
    for(i=vborder+height-1;i<vborder+height+vborder;i++)
      setInteger(target,i,0,getInteger(target,i-1,0));
  }
  else {
    for(i=seam[0]+vborder;i<=height+vborder;i++)
      setInteger(target,i,0,getInteger(target,i+1,0));
  }
  
  for(j=1;j<hborder+1;j++){
    for(i=0;i<vborder+height+vborder;i++)
      setInteger(target,i,j,getInteger(target,i,j-1));
  }
  
  for(j=1;j<width;j++){
    if(seam[j]==0){
      for(i=0;i<vborder;i++)
	setInteger(target,i,j+hborder,getInteger(target,vborder+1,hborder+j));
      for(i=vborder;i<vborder+height;i++)
	setInteger(target,i,j+hborder,getInteger(target,i+1,j+hborder);
    }
    else if(seam[j]==height){
      for(i=vborder+height-1;i<vborder+height+vborder;i++)
	setInteger(target,i,j+hborder,getInteger(target,i-1,j+hborder));
    }
    else {
      for(i=seam[j]+vborder;i<=height+vborder;i++)
	setInteger(target,i,j+hborder,getInteger(target,i+1,j+hborder));
    }
  }

  for(j=0;j<hborder;j++){
    for(i=0;j<vborder+height+vborder;i++)
      setInteger(target,i,hborder+width+j,getInteger(target,i,hborder+width+j-1));
  }
  
  return 1;
  
}
#endif


  
